# ///////////////////////////////////////////////////////////////////////////
#
#  gidutils.tcl
#
# ///////////////////////////////////////////////////////////////////////////

# ///////////////////////////////////////////////////////////////////////////
#
#  namespace GiDUtils
#
# ///////////////////////////////////////////////////////////////////////////

namespace eval GiDUtils {
}

# ///////////////////////////////////////////////////////////////////////////
#
#  GiDUtils::ActivateLayer --
#
#    Active given layer. Will be created, if it does not exist
#
# ///////////////////////////////////////////////////////////////////////////

proc GiDUtils::ActivateLayer { layer } {
  if { [catch ".central.s info layer $layer"] } {
    # layer does not exists so create it
    .central.s process escape escape escape escape \
	layer new "$layer"
  } else {
    # make on & use that layer
    .central.s process escape escape escape escape \
	layer on $layer touse "$layer"
  }
}

# ///////////////////////////////////////////////////////////////////////////
#
#  ::GiDUtils::ClearLayer --
#
# ///////////////////////////////////////////////////////////////////////////

proc ::GiDUtils::ClearLayer { layer } {
  # make on the layer

  .central.s process escape escape escape escape \
      layer on $layer

  # clear all entities

  # volumes

  .central.s process escape escape escape escape \
      geometry delete volume layer:$layer escape

  # surfaces 

  .central.s process escape escape escape escape \
      geometry delete surface layer:$layer escape

  # lines

  .central.s process escape escape escape escape \
      geometry delete line layer:$layer escape

  # points

  .central.s process escape escape escape escape \
      geometry delete point layer:$layer escape escape escape escape
}

# ///////////////////////////////////////////////////////////////////////////
#
#  ::GiDUtils::SelectFromLayer -- 
#
#   Esto no funciona
#
# ///////////////////////////////////////////////////////////////////////////

proc ::GiDUtils::SelectFromLayer { layer what {filter ""} } {
  .central.s process escape escape escape escape view selectentities $what
  if { $filter != "" } {
    .central.s process filter:$filter
  }
  .central.s process layer:$layer escape escape escape escape escape
  split [string range $::GidPriv(selection) 1 end-1]
}

proc ::GiDUtils::Select { type args } {
  array set options {
    -layer ""
    -filter ""
    -multiple yes
    -ordered no
    -before ""
    -after "" 
  }
  # falta implementar -multiple -ordered
  array set options $args
  .central.s process escape escape escape escape view selectentities $type
  if { $options(-filter) ne "" } {
    .central.s process filter:$options(-filter)
  }
  if { $options(-layer) ne "" } {  
    .central.s process layer:$options(-layer) escape escape escape escape escape
  } else {
    vwait ::GidPriv(selection)
  }
  lrange [split [string trim $::GidPriv(selection) \"]] 1 end
}

# ///////////////////////////////////////////////////////////////////////////
#
#  ::GiDUtils::AllwaysNewPoint --
#
# ///////////////////////////////////////////////////////////////////////////

proc ::GiDUtils::AllwaysNewPoint { {v ""} } {
  if { $v == "" } {
    .central.s info variable CreateAlwaysNewPoint
  } else {
    .central.s process escape escape escape escape \
	utilities variables CreateAlwaysNewPoint $v escape
  }
}

# ///////////////////////////////////////////////////////////////////////////
#
#  ::GiDUtils::PickLine --
#
# ///////////////////////////////////////////////////////////////////////////

proc ::GiDUtils::PickLine { varname {end 0} } {
  if { $end } {
    upvar $varname var 
    foreach i [trace vinfo ::GidPriv(selection)] {
      trace vdelete ::GidPriv(selection) [lindex $i 0] [lindex $i 1] 
    }
    set var [eval lindex $::GidPriv(selection) 1]
  } else {
    gidmakecurrent
    .central.s process escape escape escape escape \
	selectentities justone lines
    trace var ::GidPriv(selection) w "::GiDUtils::PickLine $varname 1 ;#"
  }
}

proc ::GiDUtils::DisableGraphics { } {
  .central.s disable windows 1
  .central.s disable graphinput 1
  .central.s disable graphics 1
  .central.s disable warnline 1
}

proc ::GiDUtils::EnableGraphics { } {
  .central.s disable windows 0
  .central.s disable graphinput 0
  .central.s disable graphics 0
  .central.s disable warnline 0
}

proc ::GiDUtils::WaitState { w } {
    $w conf -cursor watch
    .central.s waitstate 1
    update
}

proc ::GiDUtils::EndWaitState { w } {
  .central.s waitstate 0
  $w conf -cursor ""
  update
}


#GiD selection accept and returns entities idetifiers (integers > 0) as
#a "compacted" list, the sintax "a:a+n" is interpreted as "a a+1 a+2 ... a+n"

#instead "4 8 9 10 11 15" compact list --> "4 8:11 15" 
#items must be positive integer numbers (can be unordered)
proc GidUtils::CompactNumberList { a } {    
    set ret ""
    set istart 0
    if { [string match "*:*" $a] } {
	#imput list is compacted
	return $a
    }
    foreach i [lsort -integer $a] {  
	if { !$istart} {
	    set istart $i
	} else {
	    if {$i != [expr $iprev+1]} {
		if { $istart != $iprev } {
		    if { [expr $istart +1] == $iprev } {
		        append ret "$istart $iprev "
		    } else { 
		        append ret "$istart:$iprev "
		    }
		} else {
		    append ret "$istart "
		}            
		set istart $i
	    }        
	}
	set iprev $i
    }
    if { $istart } {
	if { $istart != $iprev } {
	    if { [expr $istart +1] == $iprev } {
		append ret "$istart $iprev"
	    } else { 
		append ret "$istart:$iprev"
	    }
	} else {
	    append ret "$istart"
	}   
    }
    return $ret
}

#instead "4 8:11 15"  uncompact list --> "4 8 9 10 11 15"
#items must be positive increasing integer numbers
proc GidUtils::UnCompactNumberList { a } {    
    set ret ""
    foreach item $a {
	if { [regexp   {([0-9]+)[ ]*:[ ]*([0-9]+)} $item dummy istart iend] } {
	    for { set i $istart } { $i <= $iend } { incr i } {
		lappend ret $i
	    }
	} else {
	    lappend ret $item
	}
    }
    return $ret
}


proc GidUtils::CambiaColor { w colvar} {
    upvar $colvar UserColor

    set w_par $w
    if { $w == ""} {
	set w_par .gid
    }

    set color_viejo $UserColor
    set color_nuevo [ tk_chooseColor -parent $w_par -initialcolor $color_viejo]
    if { "$color_nuevo" != ""} {
	set n [ scan $color_nuevo #%2x%2x%2x r g b]
	if { $n == 3} {
	    if { $w != ""} {
		$w conf -background $color_nuevo
	    }
	    set UserColor $color_nuevo
	}
    }
}
