proc MeshErrors { {filename ""} } {
    if { $filename == "" } {
	set filename [file join [GiD_Info Project TmpDirectory] tmp-gidErrMesh]
    }
    ReadMeshErrors $filename
    CreateWindowMeshErrors
}

proc ReadMeshErrors { filename } {
    global MeshErrorsPriv
    array unset MeshErrorsPriv

    if [catch {set archivo [open $filename r]} f] {
	#WarnWinText [_ "Error: Cannot open %1\$s: %2\$s" $filename $f]
	return 1
    }
    set MeshErrorsPriv(FileName) $filename 
    #store file data in MeshErrorsPriv
    set NLines 0
    set NSurfs 0
    set NVols 0
    set NBLM 0
    set NError_AdvancingFront3D 0
    set NError_AdvancingFront2D 0
    set NError_BoundaryLayerMesh2D 0
    set CauseLine "--"
    set CauseSurf "--"
    set CauseVol "--"
    set CauseBLM "--"
    while { ![ eof $archivo ] } {
	gets $archivo aa        
	if { $aa == {[Error_Meshing_Line]} } {
	    set MeshErrorsPriv(ErrLine,$NLines,Cause) $CauseLine
	    gets $archivo aa
	    set Num 0
	    regexp {Number:[ ]*([0-9]+)} $aa {} Num
	    set MeshErrorsPriv(ErrLine,$NLines,Num) $Num
	    incr NLines
	    gets $archivo aa        
	} elseif { $aa == {[Error_Meshing_Surface]} } {
	    set MeshErrorsPriv(ErrSurf,$NSurfs,Cause) $CauseSurf
	    gets $archivo aa
	    set Num 0
	    regexp {Number:[ ]*([0-9]+)} $aa {} Num
	    set MeshErrorsPriv(ErrSurf,$NSurfs,Num) $Num
	    incr NSurfs
	    gets $archivo aa        
	} elseif { $aa == {[Error_Meshing_Volume]} } {
	    set MeshErrorsPriv(ErrVol,$NVols,Cause) $CauseVol
	    gets $archivo aa
	    set Num 0
	    regexp {Number:[ ]*([0-9]+)} $aa {} Num
	    set MeshErrorsPriv(ErrVol,$NVols,Num) $Num
	    incr NVols
	    gets $archivo aa    
	} elseif { $aa == {[Error_Meshing_SemiStructured_Volume]} } {
	    gets $archivo aa
	    regexp {Description:[ ]*(.*)} $aa {} CauseVol
	    gets $archivo aa
	    set x "Unknown"
	    set y "Unknown"
	    set z "Unknown"
	    regexp {NearPoint:[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)} $aa {} x y z
	    set MeshErrorsPriv(ErrVol,$NVols,x) $x
	    set MeshErrorsPriv(ErrVol,$NVols,y) $y
	    set MeshErrorsPriv(ErrVol,$NVols,z) $z
	    gets $archivo aa
	    #set MeshErrorsPriv(ErrVol,$NVols,Cause) $CauseVol
	    #incr NVols
	} elseif { $aa == {[Error_AdvancingFront3D]} } {           
	    gets $archivo aa
	    regexp {Description:[ ]*(.*)} $aa {} CauseVol
	    gets $archivo aa
	    set x "Unknown"
	    set y "Unknown"
	    set z "Unknown"
	    regexp {NearPoint:[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)} $aa {} x y z
	    set MeshErrorsPriv(ErrVol,$NVols,x) $x
	    set MeshErrorsPriv(ErrVol,$NVols,y) $y
	    set MeshErrorsPriv(ErrVol,$NVols,z) $z
	    gets $archivo aa
	    gets $archivo aa
	    gets $archivo aa
	    incr NError_AdvancingFront3D
	} elseif { $aa == {[Error_AdvancingFront2D]} } {           
	    gets $archivo aa
	    regexp {Description:[ ]*(.*)} $aa {} CauseSurf
	    gets $archivo aa
	    set x "Unknown"
	    set y "Unknown"
	    set z "Unknown"
	    regexp {NearPoint:[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)} $aa {} x y z
	    set MeshErrorsPriv(ErrSurf,$NSurfs,x) $x
	    set MeshErrorsPriv(ErrSurf,$NSurfs,y) $y
	    set MeshErrorsPriv(ErrSurf,$NSurfs,z) $z
	    gets $archivo aa
	    gets $archivo aa
	    gets $archivo aa
	    incr NError_AdvancingFront2D
	} elseif { $aa == {[Error_BoundaryLayerMesh2D]} } {    
	    gets $archivo aa
	    regexp {Description:[ ]*(.*)} $aa {} CauseBLM
	    gets $archivo aa
	    set x "Unknown"
	    set y "Unknown"
	    set z "Unknown"
	    regexp {NearPoint:[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)[ ]*([0-9.+eE-]+)} $aa {} x y z
	    set MeshErrorsPriv(ErrBLM,$NBLM,x) $x
	    set MeshErrorsPriv(ErrBLM,$NBLM,y) $y
	    set MeshErrorsPriv(ErrBLM,$NBLM,z) $z
	    gets $archivo aa
	    gets $archivo aa
	    gets $archivo aa
	    incr NError_BoundaryLayerMesh2D
	    incr NBLM
	} 
    }   
    close $archivo

    if { $NError_AdvancingFront2D > 0 && $NSurfs == 0 } {
	if { $NError_AdvancingFront3D == 0 } {
	    #only show this message if not exists volume errors
	    #because sometimes the surface was meshed whith other 
	    #method, and don't provide surface number 
	    set MeshErrorsPriv(ErrSurf,$NSurfs,Num) 0
	    set MeshErrorsPriv(ErrSurf,$NSurfs,Cause) $CauseSurf
	    set NSurfs 1
	}
    }

    if { $NError_AdvancingFront3D > 0 && $NVols == 0 } {
	set MeshErrorsPriv(ErrVol,$NVols,Num) 0
	set MeshErrorsPriv(ErrVol,$NVols,Cause) $CauseVol
	set NVols 1
    }

    set MeshErrorsPriv(ErrLine,NumErr) $NLines
    set MeshErrorsPriv(ErrSurf,NumErr) $NSurfs
    set MeshErrorsPriv(ErrVol,NumErr) $NVols
    set MeshErrorsPriv(ErrBLM,NumErr) $NBLM
    return 0
}

proc DestroyMeshErrors { } {
    set w ".gid.mesherrors"
    if { [winfo exists $w] } {
	destroy $w
    }
}

proc CreateWindowMeshErrors { { w ".gid.mesherrors" } } {
    global MeshErrorsPriv
    if { ![info exists MeshErrorsPriv] } {
	WarnWin [_ "Meshing error information not available"]
	return 1
    }
    InitWindow $w [_ "MeshErrors"] MeshErrorsWindowGeom
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0

    frame $w.fup
    
    set NSurfs $MeshErrorsPriv(ErrSurf,NumErr)
    set NVols $MeshErrorsPriv(ErrVol,NumErr)
    set NBLM $MeshErrorsPriv(ErrBLM,NumErr)
    set def_back [$w cget -background]

    set fSurfaces ""
    set fVolumes ""
    set fBLMeshes ""

    NoteBook $w.fup.opt -internalborderwidth 1 -activebackground [CCColorActivo $def_back]

    
    if { $NSurfs > 0 } {
	set fSurfaces [$w.fup.opt insert end Surfaces -text [_ "Surfaces"] \
		-raisecmd {set ::MeshErrorsPriv(CurrentType) ErrSurf}] 
	CreateWidgetsFrame $fSurfaces ErrSurf 
	$w.fup.opt raise Surfaces
    }
    
    if { $NVols > 0 } {
	set fVolumes [$w.fup.opt insert end Volumes -text [_ "Volumes"] \
		-raisecmd {set ::MeshErrorsPriv(CurrentType) ErrVol}] 
	CreateWidgetsFrame $fVolumes ErrVol
	if { $NSurfs == 0 } {
	    $w.fup.opt raise Volumes
	}
    }
    
    if { $NBLM > 0 } {
	set fBLMeshes [$w.fup.opt insert end BLMeshes -text [_ "Boundary Layer"] \
		-raisecmd {set ::MeshErrorsPriv(CurrentType) ErrBLM}] 
	CreateWidgetsFrame $fBLMeshes ErrBLM
	if { $NSurfs == 0  && $NVols == 0 } {
	    $w.fup.opt raise BLMeshes
	}
    }

    grid $w.fup -sticky nsew
    pack $w.fup.opt -fill both -expand yes -side top
    update idletasks

    #lower buttons
    frame $w.frmButtons -bg [CCColorActivo $def_back]

    button $w.frmButtons.btnsignal -text [_ "Signal"] \
	    -command [list SignalSelected2 $fSurfaces.frtxt.t $fVolumes.frtxt.t] \
	    -underline 0
    button $w.frmButtons.btnclose -text [_ "Close"] \
	    -command "destroy $w" \
	    -underline 0
    SetWidgetsWidthFromText $w.frmButtons.btnsignal $w.frmButtons.btnclose
    
    grid $w.frmButtons -sticky ews -columnspan 7
    if { $::tcl_version >= 8.5 } { grid anchor $w.frmButtons center }
    grid $w.frmButtons.btnsignal $w.frmButtons.btnclose -padx 3 -pady 3
    grid columnconf $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    focus $w.frmButtons.btnclose
    bind $w <Alt-c> "tkButtonInvoke $w.frmButtons.btnclose"
    bind $w <Escape> "tkButtonInvoke $w.frmButtons.btnclose"

    wm minsize $w 160 200

    update idletasks
    
    #fill values
    if { $fSurfaces != "" } {
	$fSurfaces.lbl configure -text [_ "%s wrong surfaces" $NSurfs]
	$fSurfaces.frtxt.t del 0 end
	$fSurfaces.frtxt.t insert end [_ "Num"][string repeat " " 7][_ "Description"]
	for {set i 0} {$i < $NSurfs } {incr i} {
	    if { $MeshErrorsPriv(ErrSurf,$i,Num) > 0 } {
		set buf "$MeshErrorsPriv(ErrSurf,$i,Num)   "
	    } else {
		set buf "?   "
	    }
	    if { $MeshErrorsPriv(ErrSurf,$i,Cause) == "Couldn't map this point" } {
		if { $MeshErrorsPriv(ErrSurf,$i,Num) > 0 } {
		    set intersect [GiD_Info check surface $MeshErrorsPriv(ErrSurf,$i,Num) selfintersection]
		} else {
		    set intersect ""
		}
		if { $intersect != "" } {
		    append buf [_ "Detected %s self intersections." [llength $intersect]]  
		} else {
		    append buf [_ $MeshErrorsPriv(ErrSurf,$i,Cause)]
		}   
	    } elseif { $MeshErrorsPriv(ErrSurf,$i,Cause) == "Failed entering octtree" } {
		if { $MeshErrorsPriv(ErrSurf,$i,Num) > 0 } {
		    set intersect [GiD_Info check surface $MeshErrorsPriv(ErrSurf,$i,Num) selfintersection]
		} else {
		    set intersect ""
		}
		if { $intersect != "" } {
		    append buf [_ "Detected %s self intersections." [llength $intersect]]  
		} else {
		    append buf [_ "Check orientation"]
		}   
	    } elseif { $MeshErrorsPriv(ErrSurf,$i,Cause) == "Failed find close faces to new node" || \
		    $MeshErrorsPriv(ErrSurf,$i,Cause) == "Failed calc position of new node" } {    

		if { $MeshErrorsPriv(ErrSurf,$i,Num) > 0 } {
		    set isclosed [GiD_Info check surface $MeshErrorsPriv(ErrSurf,$i,Num) isclosed]
		} else {
		    set isclosed -1
		}
		#set isdegenerated [GiD_Info check surface $MeshErrorsPriv(ErrSurf,$i,Num) isdegeneratedboundary]
		if { $isclosed > 0 } {
		    if { $isclosed == 1 } {
		        set sense "u"
		    } elseif { $isclosed == 2 } {
		        set sense "v"
		    } else {
		        set sense "u,v"
		    }
		    append buf [_ "closed surface: Try to divide in %s" $sense]
		} else {
		    append buf [_ $MeshErrorsPriv(ErrSurf,$i,Cause)]
		}      
	    } else {
		append buf [_ $MeshErrorsPriv(ErrSurf,$i,Cause)]
	    }
	    $fSurfaces.frtxt.t insert end $buf
	}
	bind $fSurfaces.frtxt.t <Double-Button-1> [list SignalSelected $fSurfaces.frtxt.t]
    }
    
    if { $fVolumes != "" } {
	$fVolumes.lbl configure -text [_ "%s wrong volumes" $NVols]
	$fVolumes.frtxt.t del 0 end
	$fVolumes.frtxt.t insert end [_ "Num"][string repeat " " 7][_ "Description"]
	for {set i 0} {$i < $NVols } {incr i} {
	    if { $MeshErrorsPriv(ErrVol,$i,Num) > 0 } {
		set buf "$MeshErrorsPriv(ErrVol,$i,Num)    "
	    } else {
		set buf "?    "
	    }
	    if { $MeshErrorsPriv(ErrVol,$i,Cause) == "Couldn't mesh: not enough space in octtree -2-" } {
		if { $MeshErrorsPriv(ErrVol,$i,Num) > 0 } {
		    set bad [GiD_Info check volume $MeshErrorsPriv(ErrVol,$i,Num) orientation]
		} else {
		    set bad 0
		}
		if { [lindex $bad 0] > 0 } {
		    append buf [_ "%s bad oriented surfaces: %s. Use repair to correct it" \
		            [lindex $bad 0] [lindex $bad 1]]
		} else {
		    append buf [_ "Cause unknown. Use repair to try to correct it"]
		}
	    } else {
		#example SencilloPeroNoMalla: "Couldn't mesh at this location."
		append buf [_ $MeshErrorsPriv(ErrVol,$i,Cause)]
	    }
	    $fVolumes.frtxt.t insert end $buf
	}
	bind $fVolumes.frtxt.t <Double-Button-1> [list SignalSelected $fVolumes.frtxt.t]
    }

    if { $fBLMeshes != "" } {
	$fBLMeshes.lbl configure -text [_ "Boundary layer mesh error" ]
	$fBLMeshes.frtxt.t del 0 end
	$fBLMeshes.frtxt.t insert end [_ "Num"][string repeat " " 7][_ "Description"]
	for {set i 0} {$i < $NVols } {incr i} {
	    if { $MeshErrorsPriv(ErrBLM,$i,Num) > 0 } {
		set buf "$MeshErrorsPriv(ErrBLM,$i,Num)    "
	    } else {
		set buf "?    "
	    }
	    
	    #example SencilloPeroNoMalla: "Couldn't mesh at this location."
	    append buf [_ $MeshErrorsPriv(ErrBLM,$i,Cause)]
	    
	    $fBLMeshes.frtxt.t insert end $buf
	}
	bind $fBLMeshes.frtxt.t <Double-Button-1> [list SignalSelected $fBLMeshes.frtxt.t]
    }

    return 0
}


proc SignalSelected2 { w1 w2 } {
    global MeshErrorsPriv
    if { $MeshErrorsPriv(CurrentType) == "ErrSurf" } {
	SignalSelected $w1
    } elseif { $MeshErrorsPriv(CurrentType) == "ErrVol" } {
	SignalSelected $w2
    } else {
	WarnWin [_ "Must select the error line to point near the problematic region"]
    }
}


proc SignalSelected { w } {
    global MeshErrorsPriv
    set i [expr [$w curselection]-1]
    
    set type $MeshErrorsPriv(CurrentType)
    set NumOrNearPoint NearPoint
    if { $i < 0  || $i >= $MeshErrorsPriv($type,NumErr) } return
  
    set iNumEnt $MeshErrorsPriv($type,$i,Num)
    if { $type == "ErrSurf" } {
	if { $NumOrNearPoint == "NearPoint" } {
	    if { [info exists MeshErrorsPriv($type,$i,x)] } {
		set x $MeshErrorsPriv($type,$i,x)
		set y $MeshErrorsPriv($type,$i,y)
		set z $MeshErrorsPriv($type,$i,z)
	    } else {
		set x Unknown
	    }
	    if { $x == "Unknown" } {
		WarnWin [_ "Unknown point near the problematic shape"]
	    } else {
		GiD_Process Mescape Utilities SignalEntities Points FNoJoin $x $y $z
	    }
	} else {
	    GiD_Process Mescape Utilities SignalEntities Surfaces $iNumEnt
	}
    } elseif {$type == "ErrVol"} {
	if { $NumOrNearPoint == "NearPoint" } {
	    if { [info exists MeshErrorsPriv($type,$i,x)] } {
		set x $MeshErrorsPriv($type,$i,x)
		set y $MeshErrorsPriv($type,$i,y)
		set z $MeshErrorsPriv($type,$i,z)
	    } else {
		set x Unknown
	    }
	    if { $x == "Unknown" } {
		WarnWin [_ "Unknown point near the problematic shape"]
	    } else {
		GiD_Process Mescape Utilities SignalEntities Points FNoJoin $x $y $z
	    }
	} else {
	    GiD_Process Mescape Utilities SignalEntities Volumes $iNumEnt
	}
    } elseif {$type == "ErrBLM"} {
#         if { $NumOrNearPoint == "NearPoint" } {
  	    if { [info exists MeshErrorsPriv($type,$i,x)] } {
		set x $MeshErrorsPriv($type,$i,x)
		set y $MeshErrorsPriv($type,$i,y)
		set z $MeshErrorsPriv($type,$i,z)
	    } else {
		set x Unknown
	    }
	    if { $x == "Unknown" } {
		WarnWin [_ "Unknown point near the problematic shape"]
	    } else {
		GiD_Process Mescape Utilities SignalEntities Points FNoJoin $x $y $z
	    }
#         } else {
#             GiD_Process Mescape Utilities SignalEntities Volumes $iNumEnt
#         }
    }
}


proc CreateWidgetsFrame { fr type } {
    label $fr.lbl -text [_ "Number of wrong entities"]:
    frame $fr.frtxt -bd 0
   
    grid $fr.lbl -sticky new
    grid $fr.frtxt -sticky nsew -columnspan 7
    
    grid rowconf $fr 1 -weight 1
    grid columnconf $fr 0 -weight 1

    scrollbar $fr.frtxt.scrolly -command "$fr.frtxt.t yview" -orient vertical -relief sunken
    scrollbar $fr.frtxt.scrollx -command "$fr.frtxt.t xview" -orient horizontal -relief sunken
    listbox $fr.frtxt.t -font SmallFont -yscrollcommand "$fr.frtxt.scrolly set" \
	    -xscrollcommand "$fr.frtxt.scrollx set" -width 40 -height 5
    grid $fr.frtxt.t -row 0 -column 0 -sticky nsew
    grid $fr.frtxt.scrolly -row 0 -column 1 -sticky ns
    grid $fr.frtxt.scrollx -row 1 -column 0 -sticky ew
    grid rowconf $fr.frtxt 0 -weight 1
    grid columnconf $fr.frtxt 0 -weight 1

    grid remove $fr.frtxt.scrolly
    grid remove $fr.frtxt.scrollx
    bind $fr.frtxt.t <Configure> "ShowOrHideScroll $fr.frtxt.t $fr.frtxt.scrollx $fr.frtxt.scrolly"
    bind $fr.frtxt.t <3> [list ContextualMeshErrors $fr.frtxt.t %x %y %X %Y]
}

proc ContextualMeshErrors { lb x y X Y} {
    #if { ![$lb selection includes "@$x,$y"] } {
	#$lb selection clear 0 end
	#$lb selection set "@$x,$y"
    #}
    if { [$lb curselection] != [$lb index @$x,$y] } {
	$lb selection clear 0 end
	$lb selection set "@$x,$y"
    }


    set menu $lb.menu
    destroy $menu
    menu $menu -tearoff 0
    
    $menu add command -label [_ "Signal problematic point"] -command [list SignalSelected $lb]
    $menu add command -label [_ "More help"]... -command [list MeshErrorsHelp $lb]
    $menu add separator
    $menu add command -label [_ "List"]... -command [list Details]

    tk_popup $menu $X $Y
}

#text must be a listbox, text or canvas widget and sx,sy scrollbar widgets
proc ShowOrHideScroll { text sx sy } {
    foreach i "x y" {
	if { ![info exists s${i}] || ![winfo exists [set s${i}]] } { continue }
	foreach "${i}1 ${i}2" [$text ${i}view] break
	
	if { [set ${i}1] == 0 && [set ${i}2] == 1 } {
	    after idle grid remove [set s${i}]
	} else {
	    after idle grid [set s${i}]
	}
    }
}

proc Details { } {    
    global MeshErrorsPriv
    set text ""
    set NSurfs $MeshErrorsPriv(ErrSurf,NumErr)
    if { $NSurfs > 0 } {
	append text "$NSurfs problematic surfaces:"
	for {set i 0} {$i < $MeshErrorsPriv(ErrSurf,NumErr) } {incr i} {
	    append text " $MeshErrorsPriv(ErrSurf,$i,Num)"
	}
	append text "\n"
    }
    set NVols $MeshErrorsPriv(ErrVol,NumErr)
    if { $NVols > 0 } {
	append text "$NVols problematic volumes:"
	for {set i 0} {$i < $MeshErrorsPriv(ErrVol,NumErr) } {incr i} {
	    append text " $MeshErrorsPriv(ErrVol,$i,Num)"
	}
	append text "\n"
    }     
    WarnWinText $text [_ "Problems meshing"]
}

proc MeshErrorsHelp { lb } {
    global MeshErrorsPriv
    if { [$lb curselection] == "" } return
    #WarnWinText "show more inforation and advices about the error  [$lb get [$lb curselection]]"
    #set info  [$lb get [$lb curselection]]
    #set i [lindex $info 0]
    #set description [lrange $info 1 end]

    set i [expr [$lb curselection]-1]
    set type $MeshErrorsPriv(CurrentType)
    if { $i < 0  || $i >= $MeshErrorsPriv($type,NumErr) } return

    if { $type == "ErrSurf" } {
	set text [_ "Surface number %d" $MeshErrorsPriv($type,$i,Num)]
	switch -regexp $MeshErrorsPriv($type,$i,Cause) {
	    "Couldn't map this point" {
		#"Check self intersections
		append text \n[_ "Check if the boundaries have self intersections."]
	    }
	    "Failed entering octtree" {
		#"Check self intersections, or bad oriented boundary"
		append text \n[_ "Check if the boundaries have self intersections."]
		append text \n[_ "Also verify if the surface normal is agree with its boundary lines sense (right hand rule)"]
	    }
	    "Failed find close faces to new node" {
		#"Check self intersections, or bad oriented boundary"
		append text \n[_ "Check if boundaries have self intersections or some very small corner angle."]
	    }
	    "Failed FillInitialFaces" {
		set area 1
		regexp {Area=([^\n]+)\n} [GiD_Info list_entities -more surfaces $MeshErrorsPriv($type,$i,Num)] {} area
		if { $area <=  1e-25 } {
		    append text  \n[_ "The surface has zero area, probably must be deleted. Can try to collapse the model with a small tolerence"]
		} else {
		    if { [GiD_Set AutomaticCorrectSizes] == 0 } {
		        append text  \n[_ "The preference 'Automatic correct sizes' is unselected: must accurately assign mesh sizes"]
		    } else {
		        append text  \n[_ "Check the assignment of mesh sizes"]
		    }
		}
	    }
	    "Failed calc position of new node" {
		if { [GiD_Set AutomaticCorrectSizes] == 0 } {
		    append text  \n[_ "The preference 'Automatic correct sizes' is unselected: must accurately assign mesh sizes"]
		} else {
		    append text  \n[_ "Check the assignment of mesh sizes"]
		}
	    }
	    default {
		append text  \n[_ "Not available more information about the error '%s'" $MeshErrorsPriv($type,$i,Cause)]
	    }
	}
	set gentext [_ "General advices to generate the surface mesh"]:
	append gentext \n-[_ "Avoid corners with very small angles"]
	append gentext \n-[_ "Divide in two parts closed surfaces: e.g. instead a cylinder make two semi-cylinders"]
	append gentext \n-[_ "Try to delete and create again from his boundaries the problematic surface"]
	if { [GiD_Set AutomaticCorrectSizes] == 0 } {
	    append gentext \n-[_ "When the preference 'Automatic correct sizes' is unselected, must manually assign appropiated mesh sizes to geometric entities"]
	}
	append text "\n\n$gentext"
    } elseif { $type == "ErrVol" } {
	set text [_ "Volume number %d" $MeshErrorsPriv($type,$i,Num)]
	switch -regexp $MeshErrorsPriv($type,$i,Cause) {
	    "Contour element ([0-9])+ is not correct" {
		#"Contour element 126 is not correct"
		#ejemplo: caso de un cono con una sola superficie de revolucion de 360 grados
		append text \n[_ "If the boudary surface is closed, try to divide it in two parts"]
	    }
	    "Couldn't mesh: not enough space in octtree -2-" {
		append text \n[_ "If the boundary surface is closed try to divide it in two parts: e.g. instead a cylinder make two semi-cylinders"]
	    }
	    default {
		append text \n[_ "Not available more information about the error '%s'" $MeshErrorsPriv($type,$i,Cause)]
		
	    }
	}
	set gentext [_ "General advices to generate the volume mesh"]:
	append gentext \n[_ "1- Force the mesh generation for the volume surfaces (Mesh criteria->Mesh->Surfaces)"]
	append gentext \n[_ "2- See this surface mesh near the problematic point."]
	append gentext \n[_ "Typical problems are element intersections, small angles, big elements near small ones, etc."]
	append gentext \n[_ "3- Modify the geometry, or the assignment of mesh sizes to solve the problem."]
	if { [GiD_Set AutomaticCorrectSizes] == 0 } {
	    append gentext \n-[_ "When the preference 'Automatic correct sizes' is unselected, must manually assign appropiated mesh sizes to geometric entities"]
	}
	append text "\n\n$gentext"
    } else {
	set text [_ "Unexpected entity type %s" $type]
    }
    WarnWinText $text
}

#***** surface errors rsurf *****
#"Face is not correct"
#"failed in: FindClosePointsSur"
#"Couldn't map this point"
#"Couldn't mesh at this location."
#"Error entering octree"
#"User stop"
#"Couldn't map point in quadratic"
#"Advancing front has grown too much"
#"-NNN-"

#***** surface errors rfast *****
#"Failed FillInitialFaces"
#"Failed calc position of new node"
#"Failed find close faces to new node"
#"Failed entering octtree"
#"Failed remove face"
#"Failed create face"
#"Failed"

#***** volume errors *****
#"Failed entering octtree -1-"
#"Couldn't mesh at this location."
#"Couldn't mesh: not enough space in octtree -2-"
#"Error TakeFaceOut face -1-"
#"Error TakeFaceOut face -2-"
#"Error creating face -1-"
#"-NNN-"

#***** delaunay  *****
#"Error entering quadtree"
#"Error finding elements for face around"
#"Error finding elements around"

#***** semistructured *****
#"Failed entering octtree"
#"Failed meshing volume lines"
#"Failed meshing Master volume surface"
#"Failed meshing lateral structured surfaces"
#"Failed deforming volume reference mesh"
#"Failed creating interior nodes of volume reference mesh"
#"Failed creating elements of volume reference mesh"
#"Mesh already exists"
#"Failed sorting lateral surfaces nodes"
#"Failed meshing volume surfaces"
#"Failed meshing Slave volume surface"
#"Volume surfaces meshes haven't quality enough to mesh the volume."
#"Failed splitting to tetrahedra mesh"
#"Failed meshing Semistructured volume"
