
proc PickPointInSetCoord { } {
    global SetCoordPriv
    set p1 [::GidUtils::GetCoordinates [_ "Enter point (ESC to leave)"]]
    if { $p1 != "" } {
	set SetCoordPriv(xrel) [lindex $p1 0]
	set SetCoordPriv(yrel) [lindex $p1 1]
	set SetCoordPriv(zrel) [lindex $p1 2]
    }
}

proc RelativeCenterWinClose { w } {
    global SetCoordPriv
    set axes { x y z }
    foreach i $axes {
	set number $SetCoordPriv(${i}rel)
	if { ![string is double $number] } {
	    WarnWin [_ "value '%s' is not a number" $number]
	    focus $w.${i}e
	    return
	}
    }
    destroy $w
}

proc ChangeCoordLA { w } {
    global SetCoordPriv

    set w $w.relcen

    if { $SetCoordPriv(CoordTypeLA) == "-RelativeCenter-" } {
	InitWindow $w [_ "Relative center"] SetCoordRelCenterWindowGeom "" "" 1
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0

	frame $w.frmCoord
	set axes { x y z }
	foreach i $axes {
	    if { ![info exists SetCoordPriv(${i}rel)] } {
		set SetCoordPriv(${i}rel) 0.00000
	    }
	    label $w.${i}l -text $i:
	    entry $w.${i}e -relief sunken -width 20 -textvariable \
		    SetCoordPriv(${i}rel)
	    bind $w.${i}e <FocusIn> { 
		if { [%W get] != "" } {
		    %W sel from 0 
		    %W sel to end
		}
	    }
	    bind $w.${i}e <FocusOut> { %W sel clear}
	}

	set def_back [$w cget -background]
	frame $w.frmButtons -bg [CCColorActivo $def_back]

	button $w.frmButtons.btnpick -text [_ "Pick"] \
		-command "PickPointInSetCoord" \
		-underline 0 -takefocus 0
	
	button $w.frmButtons.btnclose -text [_ "Close"] \
		-command "RelativeCenterWinClose $w" \
		-underline 0 -takefocus 0
	SetWidgetsWidthFromText $w.frmButtons.btnpick $w.frmButtons.btnclose

	grid $w.xl $w.xe -in $w.frmCoord -sticky ew
	grid $w.yl $w.ye -in $w.frmCoord -sticky ew
	grid $w.zl $w.ze -in $w.frmCoord -sticky ew
	grid columnconf $w.frmCoord "1" -weight 1

	grid $w.frmCoord -sticky ew -padx 3 -pady 3
	grid $w.frmButtons -sticky ews -columnspan 7
	if { $::tcl_version >= 8.5 } { grid anchor $w.frmButtons center }
	grid $w.frmButtons.btnpick $w.frmButtons.btnclose -padx 5 -pady 6
	grid columnconf $w "0" -weight 1
	grid rowconfigure $w "1" -weight 1

	focus $w.xe

	bind $w <Return> "tkButtonInvoke $w.frmButtons.btnpick"
	bind $w <Alt-c> "tkButtonInvoke $w.frmButtons.btnclose"
	bind $w <Escape> "tkButtonInvoke $w.frmButtons.btnclose"

    } elseif { $SetCoordPriv(CoordTypeLA) == "-DefineNew-" } {
	GiD_Process escape escape escape escape data LocalAxes DefineLocAxes
	set SetCoordPriv(CoordTypeLA) "-Global-"
    }
}

proc ChangeCoordType { w } {
    global SetCoordPriv
    switch $SetCoordPriv(CoordType) {
	Cartesian {
	    $w.xl configure -text x:
	    $w.yl configure -text y:
	    $w.zl configure -text z:
	}
	Cilindrical {
	    $w.xl configure -text r:
	    $w.yl configure -text theta:
	    $w.zl configure -text z:
	}
	Spherical {
	    $w.xl configure -text r:
	    $w.yl configure -text theta:
	    $w.zl configure -text phi:
	}
    }
}


proc ApplyCmd { w } {
    global SetCoordPriv
    
    set axes { x y z }
    set retval ""
    if { $SetCoordPriv(CoordTypeLA) == "-RelativeLast-" || \
	$SetCoordPriv(CoordTypeLA) == "-RelativeCenter-" } {
	append retval @
    }
    if { ![regexp {^-[^-]*-$} $SetCoordPriv(CoordTypeLA)] } {
	append retval '$SetCoordPriv(CoordTypeLA)'
    }

    foreach i $axes {
	set number [string trim [$w.${i}e get]]
	if { ![string is double $number] } {
	    WarnWin [_ "value '%s' is not a number" $number]
	    focus $w.${i}e
	    return
	}
	append retval $number
	if { $i == "x" } {
	    if { $SetCoordPriv(CoordType) == "Cartesian" } {
		append retval ,
	    } else { 
		append retval < 
	    }
	} elseif { $i == "y" } {
	    if { $SetCoordPriv(CoordType) == "Spherical"} {
		append retval <
	    } else {
		append retval ,
	    }
	}
    }

    if { $SetCoordPriv(CoordTypeLA) == "-RelativeCenter-" } {
	set retvalLA ""
	foreach i $axes {
	    set number $SetCoordPriv(${i}rel)
	    if { ![string is double $number] } {
		WarnWin [_ "value '%s' in local center is not a number" $number]
		catch { focus $w.relcen.${i}e }
		return
	    }            
	    append retvalLA $number
	    if { $i != "z" } { append retvalLA , }
	}
	GiD_Process options EnterRelative $retvalLA FNoJoin $retval
    } else { 
	GiD_Process FNoJoin $retval
    }

    set oldtext [$w.message cget -text]
    if { $oldtext != [_ "Entered one new point"] } {
	$w.message configure -text [_ "Entered one new point"]
	after 1000 "$w.message configure -text [list $oldtext]"
    }
}


proc UpdateSetCoord { } {
    if { ![winfo exists .gid.wSetCoord] } {
	return
    }

    global SetCoordPriv
    set labels [list [_ "Global"] [_ "Relative center"]... [_ "Relative last"] [_ "Define new"]...] 
    set values {-Global- -RelativeCenter- -RelativeLast- -DefineNew-}
    foreach i [GiD_Info localaxes] {
	lappend labels $i
	lappend values $i
    }
    if { [lsearch $values $SetCoordPriv(CoordTypeLA)] == "-1" } { 
	set SetCoordPriv(CoordTypeLA) -Global- 
    }
    ComboBoxRAM $SetCoordPriv(LAcombobx) SetCoordPriv(CoordTypeLA) $labels $values

    set input [GiD_Set CreateAlwaysNewPoint]
    if { $input == 1 } { 
	set SetCoordPriv(CreateNewPoint) [_ "Always"]
    } elseif { $input == 0 } { 
	set SetCoordPriv(CreateNewPoint) [_ "Ask"] 
    } elseif { $input == -2 } { 
	set SetCoordPriv(CreateNewPoint) [_ "Never"]
    } else {
	set SetCoordPriv(CreateNewPoint) [_ "No"] 
    }
}

proc SetCoordEnable { what { w .gid.wSetCoord } } {
    global SetCoordPriv
    if { $what == "yes" } {
	set SetCoordPriv(state) normal
    } else {
	set SetCoordPriv(state) disabled
    }
    if { [winfo exists $w] } {
	$w.frmButtons.btnApply configure -state $SetCoordPriv(state)
    }
}

proc SetCoord { { w .gid.wSetCoord } } {
    global SetCoordPriv

    InitWindow $w [_ "Coordinates window"] SetCoordWindowGeom SetCoord
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0

    set def_back [$w cget -background]
    
    frame $w.frmButtons -bg [CCColorActivo $def_back]

    if { ![info exists SetCoordPriv(state)] } {
	set SetCoordPriv(state) disabled
    }
    button $w.frmButtons.btnApply -text [_ "Apply"] \
	    -command "ApplyCmd $w" \
	    -underline 0 -width 6 -state $SetCoordPriv(state) -takefocus 0
   
    button $w.frmButtons.btnclose -text [_ "Close"] \
	    -command "GiD_Process escape ;  destroy $w" \
	    -underline 0 -width 6 -takefocus 0

    SetWidgetsWidthFromText $w.frmButtons.btnApply $w.frmButtons.btnclose

    label $w.csl -text [_ "C. System"]:
    set SetCoordPriv(LAcombobx) [ComboBoxRAM $w.csla SetCoordPriv(CoordTypeLA) \
	    [list [_ "Global"] [_ "Relative center"]... [_ "Relative last"] [_ "Define new"]...] \
	    {-Global- -RelativeCenter- -RelativeLast- -DefineNew-}]

    $w.csla configure -width 12 -highlightthickness 0

    label $w.lae -text [_ "Local axes"]:
    ComboBoxRAM $w.cs SetCoordPriv(CoordType) \
	    [list [_ "Cartesian"] [_ "Cilindrical"] [_ "Spherical"]] \
	    {Cartesian Cilindrical Spherical}

    frame $w.coord

    set axes { x y z }
    foreach i $axes {
	label $w.${i}l -text $i:
	entry $w.${i}e -relief sunken -width 20 -textvariable \
		SetCoordPriv(${i}axe)
	$w.${i}e del 0 end
	$w.${i}e ins end 0.00000
	bind $w.${i}e <FocusIn> { 
	    if { [%W get] != "" } {
		%W sel from 0 
		%W sel to end
	    }
	}
	bind $w.${i}e <FocusOut> { %W sel clear}
	bind $w.${i}e <Return> "$w.frmButtons.btnApply configure -relief sunken ; \
		update idletasks ; $w.frmButtons.btnApply invoke ; $w.frmButtons.btnApply configure \
		-relief raised"
    }

    frame $w.newpnt
    label $w.newpntl1 -text [_ "Create new point"]:
    label $w.newpntl2 -relief ridge -bd 2 -textvar SetCoordPriv(CreateNewPoint)
    button $w.newpntb -text [_ "Change"] -padx 1 \
	    -command {ChangeVariables .gid.wvar General ; focus .gid.wvar.f.nb.fGeneral.menubutton#1} \
	    -takefocus 0

    label $w.message -text [_ "Use 'tab Shift-tab and Return'"] -relief ridge -bd 2

    grid $w.newpntl1 $w.newpntl2 $w.newpntb -in $w.newpnt -padx 2 -pady 2
    grid configure $w.newpntl2 -ipadx 2 -ipady 2

    grid $w.csl $w.cs -sticky e -padx 2 -pady 2
    grid  $w.lae $w.csla -sticky e -padx 2 -pady 2

    grid $w.xl $w.xe -in $w.coord -sticky ew -padx 2 -pady 2
    grid configure $w.xl -sticky w
    grid $w.yl $w.ye -in $w.coord -sticky ew -padx 2 -pady 2
    grid configure $w.yl -sticky w
    grid $w.zl $w.ze -in $w.coord -sticky ew -padx 2 -pady 2
    grid configure $w.zl -sticky w
    grid columnconf $w.coord "1" -weight 1

    grid $w.coord -sticky ew -columnspan 7
    grid $w.newpnt -sticky e -columnspan 7
    grid $w.message -sticky ews -columnspan 7

    grid $w.frmButtons -sticky ews -columnspan 7
    if { $::tcl_version >= 8.5 } { grid anchor $w.frmButtons center }
    grid $w.frmButtons.btnApply $w.frmButtons.btnclose -padx 5 -pady 6
    grid columnconf $w "0" -weight 1
    grid rowconfigure $w "4" -weight 1

    wm minsize $w 175 200

    focus $w.xe

    foreach i [bind .gid] {
	bind $w $i [bind .gid $i]
    }
    bind $w <Alt-c> [list tkButtonInvoke $w.frmButtons.btnclose]
    bind $w <Escape> [list tkButtonInvoke $w.frmButtons.btnclose]
    bind $w <Destroy> [list +DestroySetCoord %W $w] ;# + to add to previous script

    UpdateSetCoord
    ChangeCoordType $w
    trace add variable SetCoordPriv(CoordType) write "ChangeCoordType $w ;#"
    trace add variable SetCoordPriv(CoordTypeLA) write "ChangeCoordLA $w ;#"
}

proc DestroySetCoord { W w } {
    if { $W != $w } return
    global SetCoordPriv
    trace remove variable SetCoordPriv(CoordType) write "ChangeCoordType $w ;#"
    trace remove variable SetCoordPriv(CoordTypeLA) write "ChangeCoordLA $w ;#"
}
