package require snit
package require BWidget

package provide widgetsgid 1.0

################################################################################
#    ComboBoxGiD: 
#    A modified BWidget ComboBox
#    -values contain the list of language dependent showed text
#    -keyvalues contain thelist of key strings (always in english)
#    -keyvariable is an optional global variable to set or get the current key
#    -valuesvariable is an optional global variable to set/get the values
################################################################################

snit::widgetadaptor ComboBoxGiD {
    option -valuesvariable -default "" -configuremethod confvaluesvariable
    option -keyvalues -default "" -configuremethod confkeyvalues
    option -keyvariable -default "" -configuremethod confkeyvariable
    #to avoid error if try to configure -activeborderwidth -padx
    option -activeborderwidth
    option -padx
    option -editable -default "false" -configuremethod confeditable

    delegate method * to hull
    delegate option * to hull

    constructor args {
	installhull [ComboBox $self -editable $options(-editable)]
	$self configurelist $args
    }
    destructor {
	$self clear
    }
    method confgeneric {opt value} {
	set options($opt) $value
    }
    method confvaluesvariable {opt value} {
	if { $options(-valuesvariable) ne "" } {
	    trace remove variable ::$options(-valuesvariable) write \
		"[mymethod modifyvariable];#"
	}
	set options(-valuesvariable) $value

	if { $value ne "" } {
	    trace add variable ::$value write "[mymethod modifyvariable];#"
	}
    }
    method confkeyvalues {opt value} {
	set options(-keyvalues) $value
    }
    method confkeyvariable {opt value} {
	if { $options(-keyvariable) ne "" } {
	    trace remove variable ::$options(-keyvariable) write "[mymethod modifykeyvariable];#"
	}
	set options(-keyvariable) $value
	if { $value ne "" } {
	    $self configure -modifycmd "[mymethod modifyselection]"
	    trace add variable ::$value write "[mymethod modifykeyvariable];#"
	} else {
	    $self configure -modifycmd {}
	}
    }

    method confeditable { opt value} {
	set options(-editable) $value
    }

    method clear {} {
	if { $options(-valuesvariable) ne "" } {
	    trace remove variable ::$options(-valuesvariable) write \
		"[mymethod modifyvariable];#"
	}
	if { $options(-keyvariable) ne "" } {
	    trace remove variable ::$options(-keyvariable) write \
		"[mymethod modifykeyvariable];#"
	}    
    }
    method modifyvariable {} {
	upvar #0 $options(-valuesvariable) v
	if { ![info exists v] } { set v "" }
	set data [$self cget -values]
	if { $data ne $v} {
	    set prev [$self getvalue]
	    $self configure -values $v           
	    $self setvalue @$prev ;#to refresh text          
	}        
    }

    method modifyselection { } {
	upvar #0 $options(-keyvariable) v
	set i [$self getvalue]
	if { $i != -1 && [info exists options(-keyvalues)] } {
	    set v [lindex $options(-keyvalues) $i]
	} else {
	    set v ""
	}
    }
    method modifykeyvariable { } {
	upvar #0 $options(-keyvariable) v
	if { [info exists options(-keyvalues)] } {
	    set i [lsearch $options(-keyvalues) $v]
	} else {
	    set i -1
	}
	if { $i != [$self getvalue] } {
	    $self setvalue @$i
	}
    }
    #ComboBox don't support conf
    method conf { args } {
	foreach {opt val} $args {
	    if { [catch {$self cget $opt}] } {
		WarnWinText "unknown option '$opt'"
		#$self configure $opt $val
	    } else {
		$self configure $opt $val
	    }
	}
    }
}


#ComboBoxGiD .cb -values [list [_ "One"] [_ "Two"]] -keyvariable ::actual -keyvalues {One Two}
#pack .cb
#set ::actual Two
